import collections.abc
import inspect
import sys
from enum import Enum
from functools import partial
from inspect import isclass
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
    get_args,
    get_origin,
)

from cyclopts.exceptions import CoercionError
from cyclopts.utils import default_name_transform, is_union

if sys.version_info >= (3, 12):
    from typing import TypeAliasType
else:
    TypeAliasType = None

if sys.version_info < (3, 9):
    from typing_extensions import Annotated  # pragma: no cover
else:
    from typing import Annotated  # pragma: no cover


if TYPE_CHECKING:
    from cyclopts.parameter import Parameter

# from types import NoneType is available >=3.10
NoneType = type(None)
AnnotatedType = type(Annotated[int, 0])

_implicit_iterable_type_mapping: Dict[Type, Type] = {
    list: List[str],
    set: Set[str],
    tuple: Tuple[str, ...],
}

_iterable_types = {list, set}


def _bool(s: str) -> bool:
    s = s.lower()
    if s in {"no", "n", "0", "false", "f"}:
        return False
    elif s in {"yes", "y", "1", "true", "t"}:
        return True
    else:
        # Cyclopts is a little bit conservative when coercing strings into boolean.
        raise CoercionError(target_type=bool, input_value=s)


def _int(s: str) -> int:
    s = s.lower()
    if s.startswith("0x"):
        return int(s, 16)
    elif s.startswith("0b"):
        return int(s, 2)
    else:
        # Casting to a float first allows for things like "30.0"
        return int(round(float(s)))


def _bytes(s: str) -> bytes:
    return bytes(s, encoding="utf8")


def _bytearray(s: str) -> bytearray:
    return bytearray(_bytes(s))


# For types that need more logic than just invoking their type
_converters = {
    bool: _bool,
    int: _int,
    bytes: _bytes,
    bytearray: _bytearray,
}


def _convert_tuple(
    type_: Type[Any],
    *args: str,
    converter: Optional[Callable[[Type, str], Any]],
    name_transform: Callable[[str], str],
) -> Tuple:
    convert = partial(_convert, converter=converter, name_transform=name_transform)
    inner_types = tuple(x for x in get_args(type_) if x is not ...)
    inner_token_count, consume_all = token_count(type_)
    if consume_all:
        # variable-length tuple (list-like)
        remainder = len(args) % inner_token_count
        if remainder:
            raise CoercionError(
                msg=f"Incorrect number of arguments: expected multiple of {inner_token_count} but got {len(args)}."
            )
        if len(inner_types) == 1:
            inner_type = inner_types[0]
        elif len(inner_types) == 0:
            inner_type = str
        else:
            raise ValueError("A tuple must have 0 or 1 inner-types.")

        if inner_token_count == 1:
            out = tuple(convert(inner_type, x) for x in args)
        else:
            out = tuple(
                convert(inner_type, args[i : i + inner_token_count]) for i in range(0, len(args), inner_token_count)
            )
        return out
    else:
        # Fixed-length tuple
        if inner_token_count != len(args):
            raise CoercionError(msg=f"Incorrect number of arguments: expected {inner_token_count} but got {len(args)}.")
        args_per_convert = [token_count(x)[0] for x in inner_types]
        it = iter(args)
        batched = [[next(it) for _ in range(size)] for size in args_per_convert]
        batched = [elem[0] if len(elem) == 1 else elem for elem in batched]
        out = tuple(convert(inner_type, arg) for inner_type, arg in zip(inner_types, batched))
    return out


def _convert(
    type_,
    element,
    *,
    converter: Optional[Callable[[Type, str], Any]],
    name_transform: Callable[[str], str],
):
    """Inner recursive conversion function for public ``convert``.

    Parameters
    ----------
    converter: Callable
    name_transform: Callable
    """
    convert = partial(_convert, converter=converter, name_transform=name_transform)
    convert_tuple = partial(_convert_tuple, converter=converter, name_transform=name_transform)
    origin_type = get_origin(type_)
    inner_types = [resolve(x) for x in get_args(type_)]

    if type_ in _implicit_iterable_type_mapping:
        return convert(_implicit_iterable_type_mapping[type_], element)

    if origin_type is collections.abc.Iterable:
        assert len(inner_types) == 1
        return convert(List[inner_types[0]], element)  # pyright: ignore[reportGeneralTypeIssues]
    elif TypeAliasType is not None and isinstance(type_, TypeAliasType):
        return convert(type_.__value__, element)
    elif is_union(origin_type):
        for t in inner_types:
            if t is NoneType:
                continue
            try:
                return convert(t, element)
            except Exception:
                pass
        else:
            raise CoercionError(input_value=element, target_type=type_)
    elif origin_type is Literal:
        # Try coercing the token into each allowed Literal value (left-to-right).
        for choice in get_args(type_):
            try:
                res = convert(type(choice), (element))
            except Exception:
                continue
            if res == choice:
                return res
        else:
            raise CoercionError(input_value=element, target_type=type_)
    elif origin_type in _iterable_types:  # NOT including tuple
        count, _ = token_count(inner_types[0])
        if count > 1:
            gen = zip(*[iter(element)] * count)
        else:
            gen = element
        return origin_type(convert(inner_types[0], e) for e in gen)  # pyright: ignore[reportOptionalCall]
    elif origin_type is tuple:
        if isinstance(element, str):
            # E.g. Tuple[str] (Annotation: tuple containing a single string)
            return convert_tuple(type_, element, converter=converter)
        else:
            return convert_tuple(type_, *element, converter=converter)
    elif isclass(type_) and issubclass(type_, Enum):
        if converter is None:
            element_transformed = name_transform(element)
            for member in type_:
                if name_transform(member.name) == element_transformed:
                    return member
            raise CoercionError(input_value=element, target_type=type_)
        else:
            return converter(type_, element)
    else:
        # The actual casting/converting of the underlying type is performed here.
        try:
            if converter is None:
                return _converters.get(type_, type_)(element)
            else:
                return converter(type_, element)
        except ValueError:
            raise CoercionError(input_value=element, target_type=type_) from None


_unsupported_target_types = {dict}


def get_origin_and_validate(type_: Type):
    origin_type = get_origin(type_)
    if origin_type is None:
        if type_ in _unsupported_target_types:
            raise TypeError(f"Unsupported Type: {type_}")
    else:
        if origin_type in _unsupported_target_types:
            raise TypeError(f"Unsupported Type: {type_}")
    return origin_type


def resolve(type_: Any) -> Type:
    """Perform all simplifying resolutions."""
    if type_ is inspect.Parameter.empty:
        return str

    type_prev = None
    while type_ != type_prev:
        type_prev = type_
        type_ = resolve_annotated(type_)
        type_ = resolve_optional(type_)
    return type_


def resolve_optional(type_: Any) -> Type:
    """Only resolves Union's of None + one other type (i.e. Optional)."""
    # Python will automatically flatten out nested unions when possible.
    # So we don't need to loop over resolution.

    if not is_union(get_origin(type_)):
        return type_

    non_none_types = [t for t in get_args(type_) if t is not NoneType]
    if not non_none_types:  # pragma: no cover
        # This should never happen; python simplifies:
        #    ``Union[None, None] -> NoneType``
        raise ValueError("Union type cannot be all NoneType")
    elif len(non_none_types) == 1:
        type_ = non_none_types[0]
    elif len(non_none_types) > 1:
        return Union[tuple(resolve_optional(x) for x in non_none_types)]  # pyright: ignore
    else:
        raise NotImplementedError

    return type_


def resolve_annotated(type_: Any) -> Type:
    if type(type_) is AnnotatedType:
        type_ = get_args(type_)[0]
    return type_


def convert(
    type_: Any,
    *args: str,
    converter: Optional[Callable[[Type, str], Any]] = None,
    name_transform: Optional[Callable[[str], str]] = None,
):
    """Coerce variables into a specified type.

    Internally used to coercing string CLI tokens into python builtin types.
    Externally, may be useful in a custom converter.
    See Cyclopt's automatic coercion rules :doc:`/rules`.

    If ``type_`` **is not** iterable, then each element of ``*args`` will be converted independently.
    If there is more than one element, then the return type will be a ``Tuple[type_, ...]``.
    If there is a single element, then the return type will be ``type_``.

    If ``type_`` **is** iterable, then all elements of ``*args`` will be collated.

    Parameters
    ----------
    type_: Type
        A type hint/annotation to coerce ``*args`` into.
    `*args`: str
        String tokens to coerce.
    converter: Optional[Callable[[Type, str], Any]]
        An optional function to convert tokens to the inner-most types.
        The converter should have signature:

        .. code-block:: python

            def converter(type_: type, value: str) -> Any:
                "Perform conversion of string token."

        This allows to use the :func:`convert` function to handle the the difficult task
        of traversing lists/tuples/unions/etc, while leaving the final conversion logic to
        the caller.
    name_transform: Optional[Callable[[str], str]]
        Currently only used for ``Enum`` type hints.
        A function that transforms enum names and CLI values into a normalized format.

        The function should have signature:

        .. code-block:: python

            def name_transform(s: str) -> str:
                "Perform name transform."

        where the returned value is the name to be used on the CLI.

        If ``None``, defaults to ``cyclopts.default_name_transform``.

    Returns
    -------
    Any
        Coerced version of input ``*args``.
    """
    if name_transform is None:
        name_transform = default_name_transform

    convert = partial(_convert, converter=converter, name_transform=name_transform)
    convert_tuple = partial(_convert_tuple, converter=converter, name_transform=name_transform)
    type_ = resolve(type_)

    if type_ is Any:
        type_ = str

    type_ = _implicit_iterable_type_mapping.get(type_, type_)

    origin_type = get_origin_and_validate(type_)

    if origin_type is tuple:
        return convert_tuple(type_, *args)
    elif (origin_type or type_) in _iterable_types or origin_type is collections.abc.Iterable:
        return convert(type_, args)
    elif len(args) == 1:
        return convert(type_, args[0])
    else:
        return [convert(type_, item) for item in args]


def token_count(type_: Any) -> Tuple[int, bool]:
    """The number of tokens after a keyword the parameter should consume.

    Parameters
    ----------
    type_: Type
        A type hint/annotation to infer token_count from if not explicitly specified.

    Returns
    -------
    int
        Number of tokens to consume.
    bool
        If this is ``True`` and positional, consume all remaining tokens.
        The returned number of tokens constitutes a single element of the iterable-to-be-parsed.
    """
    from cyclopts.parameter import get_hint_parameter

    annotation = get_hint_parameter(type_)[0]

    annotation = resolve(annotation)
    origin_type = get_origin_and_validate(annotation)

    if (origin_type or annotation) is tuple:
        args = get_args(annotation)
        if args:
            return sum(token_count(x)[0] for x in args if x is not ...), ... in args
        else:
            return 1, True
    elif (origin_type or annotation) is bool:
        return 0, False
    elif annotation in _iterable_types or (origin_type in _iterable_types and len(get_args(annotation)) == 0):
        return 1, True
    elif (origin_type in _iterable_types or origin_type is collections.abc.Iterable) and len(get_args(annotation)):
        return token_count(get_args(annotation)[0])[0], True
    else:
        return 1, False



import inspect
import re
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Type,
)

from attrs import define, field

import cyclopts.utils
from cyclopts.group import Group
from cyclopts.utils import ParameterDict

if TYPE_CHECKING:
    from rich.console import Console

    from cyclopts.core import App


__all__ = [
    "CoercionError",
    "CommandCollisionError",
    "CycloptsError",
    "DocstringError",
    "InvalidCommandError",
    "MissingArgumentError",
    "RepeatArgumentError",
    "UnknownOptionError",
    "UnusedCliTokensError",
    "ValidationError",
    "format_cyclopts_error",
]


def _get_function_info(func):
    return inspect.getsourcefile(func), inspect.getsourcelines(func)[1]


class CommandCollisionError(Exception):
    """A command with the same name has already been registered to the app."""

    # This doesn't derive from CycloptsError since this is a developer error
    # rather than a runtime error.


class DocstringError(Exception):
    """The docstring either has a syntax error, or inconsistency with the function signature."""


@define(kw_only=True)
class CycloptsError(Exception):
    """Root exception for runtime errors.

    As CycloptsErrors bubble up the Cyclopts stack, more information is added to it.
    Finally, :func:`cyclopts.exceptions.format_cyclopts_error` formats the message nicely for the user.
    """

    msg: Optional[str] = None
    """
    If set, override automatic message generation.
    """

    verbose: bool = True
    """
    More verbose error messages; aimed towards developers debugging their Cyclopts app.
    Defaults to ``False``.
    """

    root_input_tokens: Optional[List[str]] = None
    """
    The parsed CLI tokens that were initially fed into the :class:`App`.
    """

    unused_tokens: Optional[List[str]] = None
    """
    Leftover tokens after parsing is complete.
    """

    target: Optional[Callable] = None
    """
    The python function associated with the command being parsed.
    """

    cli2parameter: Optional[Dict[str, Tuple[inspect.Parameter, Any]]] = None
    """
    Dictionary mapping CLI strings to python parameters.
    """

    parameter2cli: Optional[ParameterDict] = None
    """
    Dictionary mapping function parameters to possible CLI tokens.
    """

    command_chain: Optional[Iterable[str]] = None
    """
    List of command that lead to ``target``.
    """

    app: Optional["App"] = None
    """
    The Cyclopts application itself.
    """

    console: Optional["Console"] = field(default=None, kw_only=True)
    """
    Rich console to display runtime errors.
    """

    def __str__(self):
        if self.msg is not None:
            return self.msg

        strings = []
        if self.verbose:
            strings.append(type(self).__name__)
            if self.target:
                file, lineno = _get_function_info(self.target)
                strings.append(f'Function defined in file "{file}", line {lineno}:')
                strings.append(f"    {self.target.__name__}{cyclopts.utils.signature(self.target)}")
            if self.root_input_tokens is not None:
                strings.append(f"Root Input Tokens: {self.root_input_tokens}")
        else:
            pass

        if strings:
            return "\n".join(strings) + "\n"
        else:
            return ""

    def _find_and_replace(self, s: str) -> str:
        """Replaces all instances of "--python-variable-name" with "--cli-variable-name"."""
        if self.parameter2cli is None:
            return s
        for p, names in self.parameter2cli.items():
            pattern = rf"--\b{re.escape(p.name)}\b(?=\W|$)"
            replacement = names[0]
            s = re.sub(pattern, replacement, s)
        return s


@define(kw_only=True)
class ValidationError(CycloptsError):
    """Validator function raised an exception."""

    value: str
    """Parenting Assertion/Value/Type Error message."""

    parameter: Optional[inspect.Parameter] = None
    """Parameter who's ``validator`` function failed."""

    group: Optional[Group] = None
    """Group who's ``validator`` function failed."""

    def __str__(self):
        # Either parameter or group must be set (but not both!)
        assert bool(self.parameter) is not bool(self.group)

        if self.parameter:
            assert self.parameter2cli is not None
            # TODO: The displayed ``parameter_cli_name`` may not match the actual offending
            # cli --option token provided (i.e. aliases).
            # It would be much nicer to directly get the offending raw cli --option token.
            parameter_cli_name = ",".join(self.parameter2cli[self.parameter])
            return super().__str__() + f'Invalid value for "{parameter_cli_name}". {self.value}'
        elif self.group:
            # TODO: it would be much nicer to directly get the offending raw cli --option token(s).
            # However, this information is not available to the validator, so it's a bit hopeless.
            self.value = self._find_and_replace(self.value)
            return super().__str__() + f'Invalid values for group "{self.group}". {self.value}'
        else:
            raise NotImplementedError


@define(kw_only=True)
class UnknownOptionError(CycloptsError):
    """Unknown/unregistered option provided by the cli."""

    token: str

    def __str__(self):
        return super().__str__() + f'Unknown option: "{self.token}".'


@define(kw_only=True)
class CoercionError(CycloptsError):
    """There was an error performing automatic type coercion."""

    input_value: str = ""
    """
    String input token that couldn't be coerced.
    """

    target_type: Optional[Type] = None
    """
    Intended type to coerce into.
    """

    parameter: Optional[inspect.Parameter] = None

    def __str__(self):
        if self.parameter:
            assert self.parameter2cli is not None
            parameter_cli_name = ",".join(self.parameter2cli[self.parameter])

        if self.msg is not None:
            if self.parameter:
                return f"{parameter_cli_name}: " + self.msg  # pyright: ignore[reportPossiblyUnboundVariable]
            else:
                return self.msg

        response = f'Error converting value "{self.input_value}"'

        if self.target_type is not None:
            target_type = str(self.target_type).lstrip("typing.")  # lessens the verbosity a little bit.
            response += f" to {target_type}"

        if self.parameter:
            response += f' for "{parameter_cli_name}"'  # pyright: ignore[reportPossiblyUnboundVariable]

        return super().__str__() + response + "."


class InvalidCommandError(CycloptsError):
    """CLI token combination did not yield a valid command."""

    def __str__(self):
        import difflib

        assert self.unused_tokens
        token = self.unused_tokens[0]
        response = super().__str__() + f'Unable to interpret valid command from "{token}".'

        if self.app and self.app._commands:
            close_matches = difflib.get_close_matches(token, self.app._commands, n=1, cutoff=0.8)
            if close_matches:
                response += f' Did you mean "{close_matches[0]}"?'

        return response


@define(kw_only=True)
class UnusedCliTokensError(CycloptsError):
    """Not all CLI tokens were used as expected."""

    def __str__(self):
        assert self.unused_tokens is not None
        return super().__str__() + f"Unused Tokens: {self.unused_tokens}."


@define(kw_only=True)
class MissingArgumentError(CycloptsError):
    """A parameter had insufficient tokens to be populated."""

    parameter: inspect.Parameter
    """
    The parameter that failed to parse.
    """

    tokens_so_far: List[str] = field(factory=list)
    """
    The tokens that were parsed so far for this Parameter.
    """

    def __str__(self):
        from cyclopts._convert import token_count

        count, _ = token_count(self.parameter)
        if count == 0:
            required_string = "flag required"
            only_got_string = ""
        elif count == 1:
            required_string = "requires an argument"
            only_got_string = ""
        else:
            required_string = f"requires {count} arguments"
            received_count = len(self.tokens_so_far) % count
            only_got_string = f" Only got {received_count}." if received_count else ""

        assert self.parameter2cli is not None
        parameter_cli_name = ",".join(self.parameter2cli[self.parameter])

        strings = []
        if self.command_chain:
            strings.append(
                f'Command "{" ".join(self.command_chain)}" parameter "{parameter_cli_name}" {required_string}.{only_got_string}'
            )
        else:
            strings.append(f'Parameter "{parameter_cli_name}" {required_string}.{only_got_string}')

        if self.verbose:
            strings.append(f" Parsed: {self.tokens_so_far}.")

        return super().__str__() + " ".join(strings)


@define(kw_only=True)
class RepeatArgumentError(CycloptsError):
    """The same parameter has erroneously been specified multiple times."""

    parameter: inspect.Parameter
    """
    The repeated parameter.
    """

    def __str__(self):
        assert self.parameter2cli is not None
        parameter_cli_name = ",".join(self.parameter2cli[self.parameter])
        return super().__str__() + f"Parameter {parameter_cli_name} specified multiple times."


def format_cyclopts_error(e: Any):
    from rich import box
    from rich.panel import Panel
    from rich.text import Text

    panel = Panel(
        Text(str(e), "default"),
        title="Error",
        box=box.ROUNDED,
        expand=True,
        title_align="left",
        style="red",
    )
    return panel



import os
from typing import TYPE_CHECKING, Callable, Dict, List, Tuple, Union

from attrs import define, field

from cyclopts._env_var import env_var_split
from cyclopts.config import Unset

if TYPE_CHECKING:
    from cyclopts.core import App


@define
class Env:
    prefix: str = ""
    command: bool = field(default=True, kw_only=True)
    split: Callable = field(default=env_var_split, kw_only=True)

    def __call__(self, apps: List["App"], commands: Tuple[str, ...], mapping: Dict[str, Union[Unset, List[str]]]):
        prefix = self.prefix
        if self.command and commands:
            prefix += "_".join(commands) + "_"
        for key, value in mapping.items():
            if not isinstance(value, Unset) or value.related_set(mapping):
                continue
            env_key = (prefix + key).upper().replace("-", "_")
            try:
                env_value = os.environ[env_key]
            except KeyError:
                continue
            mapping[key] = self.split(value.iparam.annotation, env_value)



# Don't manually change, let poetry-dynamic-versioning handle it.
__version__ = "2.9.9"

__all__ = [
    "App",
    "CoercionError",
    "CommandCollisionError",
    "CycloptsError",
    "Dispatcher",
    "DocstringError",
    "Group",
    "InvalidCommandError",
    "MissingArgumentError",
    "Parameter",
    "UnknownOptionError",
    "UnusedCliTokensError",
    "ValidationError",
    "config",
    "convert",
    "default_name_transform",
    "env_var_split",
    "types",
    "validators",
]

from cyclopts._convert import convert
from cyclopts._env_var import env_var_split
from cyclopts.core import App
from cyclopts.exceptions import (
    CoercionError,
    CommandCollisionError,
    CycloptsError,
    DocstringError,
    InvalidCommandError,
    MissingArgumentError,
    UnknownOptionError,
    UnusedCliTokensError,
    ValidationError,
)
from cyclopts.group import Group
from cyclopts.parameter import Parameter
from cyclopts.protocols import Dispatcher
from cyclopts.utils import default_name_transform

from . import config, types, validators



from typing import Optional, Type, Union, get_args, get_origin

from attrs import frozen

Numeric = Union[int, float]


@frozen(kw_only=True)
class Number:
    """Limit input number to a value range."""

    lt: Optional[Numeric] = None
    """Input value must be **less than** this value."""

    lte: Optional[Numeric] = None
    """Input value must be **less than or equal** this value."""

    gt: Optional[Numeric] = None
    """Input value must be **greater than** this value."""

    gte: Optional[Numeric] = None
    """Input value must be **greater than or equal** this value."""

    def __call__(self, type_: Type, value: Numeric):
        origin = get_origin(type_) or type_
        if origin not in get_args(Numeric):
            raise TypeError
        if not isinstance(value, get_args(Numeric)):
            raise TypeError

        if self.lt is not None and value >= self.lt:
            raise ValueError(f"Must be < {self.lt}")

        if self.lte is not None and value > self.lte:
            raise ValueError(f"Must be <= {self.lte}")

        if self.gt is not None and value <= self.gt:
            raise ValueError(f"Must be > {self.gt}")

        if self.gte is not None and value < self.gte:
            raise ValueError(f"Must be >= {self.gte}")



import inspect
from functools import partial
from typing import (
    Any,
    Callable,
    Iterable,
    Optional,
    Tuple,
    Type,
    Union,
    cast,
    get_args,
    get_origin,
)

import attrs
from attrs import field, frozen

import cyclopts._env_var
import cyclopts.utils
from cyclopts._convert import (
    AnnotatedType,
    convert,
    get_origin_and_validate,
    resolve,
    resolve_optional,
)
from cyclopts.group import Group
from cyclopts.utils import (
    default_name_transform,
    optional_to_tuple_converter,
    record_init,
    to_tuple_converter,
)


def _double_hyphen_validator(instance, attribute, values):
    if not values:
        return

    for value in values:
        if value is not None and not value.startswith("--"):
            raise ValueError(f'{attribute.alias} value must start with "--".')


def _negative_converter(default: Tuple[str, ...]):
    def converter(value) -> Tuple[str, ...]:
        if value is None:
            return default
        else:
            return to_tuple_converter(value)

    return converter


@record_init("_provided_args")
@frozen
class Parameter:
    """Cyclopts configuration for individual function parameters."""

    # All documentation has been moved to ``docs/api.rst`` for greater control with attrs.

    # This can ONLY ever be a Tuple[str, ...]
    name: Union[None, str, Iterable[str]] = field(
        default=None,
        converter=lambda x: cast(Tuple[str, ...], to_tuple_converter(x)),
    )

    _converter: Callable = field(default=None, alias="converter")

    # This can ONLY ever be a Tuple[Callable, ...]
    validator: Union[None, Callable, Iterable[Callable]] = field(
        default=(),
        converter=lambda x: cast(Tuple[Callable, ...], to_tuple_converter(x)),
    )

    # This can ONLY ever be a Tuple[str, ...]
    negative: Union[None, str, Iterable[str]] = field(default=None, converter=optional_to_tuple_converter)

    # This can ONLY ever be a Tuple[Union[Group, str], ...]
    group: Union[None, Group, str, Iterable[Union[Group, str]]] = field(
        default=None, converter=to_tuple_converter, hash=False
    )

    parse: bool = field(default=None, converter=attrs.converters.default_if_none(True))

    _show: Optional[bool] = field(default=None, alias="show")

    show_default: Optional[bool] = field(default=None)

    show_choices: bool = field(default=None, converter=attrs.converters.default_if_none(True))

    help: Optional[str] = field(default=None)

    show_env_var: bool = field(default=None, converter=attrs.converters.default_if_none(True))

    # This can ONLY ever be a Tuple[str, ...]
    env_var: Union[None, str, Iterable[str]] = field(
        default=None,
        converter=lambda x: cast(Tuple[str, ...], to_tuple_converter(x)),
    )

    env_var_split: Callable = cyclopts._env_var.env_var_split

    # This can ONLY ever be a Tuple[str, ...]
    negative_bool: Union[None, str, Iterable[str]] = field(
        default=None,
        converter=_negative_converter(("--no-",)),
        validator=_double_hyphen_validator,
    )

    # This can ONLY ever be a Tuple[str, ...]
    negative_iterable: Union[None, str, Iterable[str]] = field(
        default=None,
        converter=_negative_converter(("--empty-",)),
        validator=_double_hyphen_validator,
    )

    required: Optional[bool] = field(default=None)

    allow_leading_hyphen: bool = field(default=False)

    name_transform: Optional[Callable[[str], str]] = field(
        default=None,
        converter=attrs.converters.default_if_none(default_name_transform),
        kw_only=True,
    )

    # Populated by the record_attrs_init_args decorator.
    _provided_args: Tuple[str] = field(default=(), init=False, eq=False)

    @property
    def show(self):
        return self._show if self._show is not None else self.parse

    @property
    def converter(self):
        return self._converter if self._converter else partial(convert, name_transform=self.name_transform)

    def get_negatives(self, type_, *names: str) -> Tuple[str, ...]:
        type_ = get_origin(type_) or type_

        if self.negative is not None:
            return self.negative  # pyright: ignore
        elif type_ not in (bool, list, set):
            return ()

        out = []
        for name in names:
            if name.startswith("--"):
                name = name[2:]
            elif name.startswith("-"):
                # Do not support automatic negation for short flags.
                continue
            else:
                raise ValueError("All parameters should have started with '-' or '--'.")

            negative_prefixes = self.negative_bool if type_ is bool else self.negative_iterable
            assert isinstance(negative_prefixes, tuple)
            for negative_prefix in negative_prefixes:
                out.append(f"{negative_prefix}{name}")
        return tuple(out)

    def __repr__(self):
        """Only shows non-default values."""
        content = ", ".join(
            [
                f"{a.alias}={getattr(self, a.name)!r}"
                for a in self.__attrs_attrs__  # pyright: ignore[reportAttributeAccessIssue]
                if a.alias in self._provided_args
            ]
        )
        return f"{type(self).__name__}({content})"

    @classmethod
    def combine(cls, *parameters: Optional["Parameter"]) -> "Parameter":
        """Returns a new Parameter with values of ``parameters``.

        Parameters
        ----------
        `*parameters`: Optional[Parameter]
             Parameters who's attributes override ``self`` attributes.
             Ordered from least-to-highest attribute priority.
        """
        kwargs = {}
        for parameter in parameters:
            if parameter is None:
                continue
            for a in parameter.__attrs_attrs__:  # pyright: ignore[reportAttributeAccessIssue]
                if a.init and a.alias in parameter._provided_args:
                    kwargs[a.alias] = getattr(parameter, a.name)

        return cls(**kwargs)

    @classmethod
    def default(cls) -> "Parameter":
        """Create a Parameter with all Cyclopts-default values.

        This is different than just :class:`Parameter` because the default
        values will be recorded and override all upstream parameter values.
        """
        return cls(
            **{a.alias: a.default for a in cls.__attrs_attrs__ if a.init}  # pyright: ignore[reportAttributeAccessIssue]
        )


def validate_command(f: Callable):
    """Validate if a function abides by Cyclopts's rules.

    Raises
    ------
    ValueError
        Function has naming or parameter/signature inconsistencies.
    """
    signature = cyclopts.utils.signature(f)
    for iparam in signature.parameters.values():
        get_origin_and_validate(iparam.annotation)
        type_, cparam = get_hint_parameter(iparam)
        if not cparam.parse and iparam.kind is not iparam.KEYWORD_ONLY:
            raise ValueError("Parameter.parse=False must be used with a KEYWORD_ONLY function parameter.")


def get_hint_parameter(type_: Any, *default_parameters: Optional[Parameter]) -> Tuple[Type, Parameter]:
    """Get the type hint and Cyclopts :class:`Parameter` from a type-hint.

    If a ``cyclopts.Parameter`` is not found, a default Parameter is returned.
    """
    cyclopts_parameters = []

    if isinstance(type_, inspect.Parameter):
        annotation = type_.annotation

        if annotation is inspect.Parameter.empty or resolve(annotation) is Any:
            if type_.default in (inspect.Parameter.empty, None):
                annotation = str
            else:
                return get_hint_parameter(type(type_.default), *default_parameters)
    else:
        annotation = type_

        if annotation is inspect.Parameter.empty:
            annotation = str

    annotation = resolve_optional(annotation)

    if type(annotation) is AnnotatedType:
        annotations = annotation.__metadata__  # pyright: ignore[reportGeneralTypeIssues]
        annotation = get_args(annotation)[0]
        cyclopts_parameters = [x for x in annotations if isinstance(x, Parameter)]
    annotation = resolve(annotation)

    cparam = Parameter.combine(*default_parameters, *cyclopts_parameters)
    return annotation, cparam



from pathlib import Path
from typing import Any, Dict

from cyclopts.config._common import ConfigFromFile


class Json(ConfigFromFile):
    def _load_config(self, path: Path) -> Dict[str, Any]:
        import json

        with path.open() as f:
            return json.load(f)



import inspect
import itertools
import os
import shlex
import sys
from contextlib import suppress
from typing import Callable, Dict, Iterable, List, Tuple, Type, Union

from cyclopts._convert import _bool, token_count
from cyclopts.config import Unset
from cyclopts.exceptions import (
    CoercionError,
    CycloptsError,
    MissingArgumentError,
    RepeatArgumentError,
    UnknownOptionError,
    ValidationError,
)
from cyclopts.parameter import get_hint_parameter, validate_command
from cyclopts.resolve import ResolvedCommand
from cyclopts.utils import ParameterDict


def normalize_tokens(tokens: Union[None, str, Iterable[str]]) -> List[str]:
    if tokens is None:
        tokens = sys.argv[1:]  # Remove the executable
    elif isinstance(tokens, str):
        tokens = shlex.split(tokens)
    else:
        tokens = list(tokens)
    return tokens


def _cli_kw_to_f_kw(cli_key: str):
    """Only used for converting unknown CLI key/value keys for ``**kwargs``."""
    assert cli_key.startswith("--")
    cli_key = cli_key[2:]  # strip off leading "--"
    cli_key = cli_key.replace("-", "_")
    return cli_key


def _parse_kw_and_flags(command: ResolvedCommand, tokens, mapping):
    kwargs_iparam = next((x for x in command.iparams if x.kind == x.VAR_KEYWORD), None)

    if kwargs_iparam:
        mapping[kwargs_iparam] = {}

    unused_tokens = []

    skip_next_iterations = 0
    for i, token in enumerate(tokens):
        # If the previous argument was a keyword, then this is its value
        if skip_next_iterations > 0:
            skip_next_iterations -= 1
            continue

        if not _is_option_like(token):
            unused_tokens.append(token)
            continue

        cli_values = []
        kwargs_key = None
        consume_count = 0

        if "=" in token:
            cli_key, cli_value = token.split("=", 1)
            cli_values.append(cli_value)
            consume_count -= 1
        else:
            cli_key = token

        try:
            iparam, implicit_value = command.cli2parameter[cli_key]
        except KeyError:
            if kwargs_iparam:
                iparam = kwargs_iparam
                kwargs_key = _cli_kw_to_f_kw(cli_key)
                implicit_value = None
            else:
                unused_tokens.append(token)
                continue

        cparam = command.iparam_to_cparam[iparam]

        if implicit_value is not None:
            # A flag was parsed
            if cli_values:
                # A value was parsed from "--key=value", and the ``value`` is in ``cli_values``.
                # Immediately convert to actual boolean datatype.
                if _bool(cli_values[-1]):
                    # --negative-flag=true or --empty-flag=true
                    cli_values[-1] = implicit_value
                else:
                    # --negative-flag=false or --empty-flag=false
                    if implicit_value in (True, False):  # This is a boolean "--no-" flag.
                        cli_values[-1] = not implicit_value
                    else:  # This is an iterable "--empty-"
                        # Just skip it, it doesn't mean anything.
                        continue
            else:
                cli_values.append(implicit_value)
            tokens_per_element, consume_all = 0, False
        else:
            tokens_per_element, consume_all = token_count(iparam)

            with suppress(IndexError):
                if consume_all:
                    for j in itertools.count():
                        token = tokens[i + 1 + j]
                        if not cparam.allow_leading_hyphen and _is_option_like(token):
                            break
                        cli_values.append(token)
                        skip_next_iterations += 1
                else:
                    consume_count += tokens_per_element
                    for j in range(consume_count):
                        token = tokens[i + 1 + j]
                        if not cparam.allow_leading_hyphen:
                            _validate_is_not_option_like(token)
                        cli_values.append(token)
                        skip_next_iterations += 1

            if not cli_values or len(cli_values) % tokens_per_element:
                raise MissingArgumentError(parameter=iparam, tokens_so_far=cli_values)

        # Update mapping
        if iparam is kwargs_iparam:
            assert kwargs_key is not None
            if kwargs_key in mapping[iparam] and not consume_all:
                raise RepeatArgumentError(parameter=iparam)
            mapping[iparam].setdefault(kwargs_key, [])
            mapping[iparam][kwargs_key].extend(cli_values)
        else:
            if iparam in mapping and not consume_all:
                raise RepeatArgumentError(parameter=iparam)

            mapping.setdefault(iparam, [])
            mapping[iparam].extend(cli_values)

    return unused_tokens


def _is_option_like(token: str) -> bool:
    """Checks if a token looks like an option.

    Namely, negative numbers are not options, but a token like ``--foo`` is.
    """
    with suppress(ValueError):
        complex(token)
        return False
    return token.startswith("-")


def _validate_is_not_option_like(token):
    if _is_option_like(token):
        raise UnknownOptionError(token=token)


def _parse_pos(
    command: ResolvedCommand,
    tokens: Iterable[str],
    mapping: ParameterDict,
) -> List[str]:
    tokens = list(tokens)

    def remaining_parameters():
        for iparam, cparam in command.iparam_to_cparam.items():
            _, consume_all = token_count(iparam)
            if iparam in mapping and not consume_all:
                continue
            if iparam.kind is iparam.KEYWORD_ONLY:  # pragma: no cover
                # the kwargs parameter should always be in mapping.
                break
            yield iparam, cparam

    for iparam, cparam in remaining_parameters():
        if not tokens:
            break

        if iparam.kind is iparam.VAR_POSITIONAL:  # ``*args``
            mapping.setdefault(iparam, [])
            for token in tokens:
                if not cparam.allow_leading_hyphen:
                    _validate_is_not_option_like(token)

                mapping[iparam].append(token)
            tokens = []
            break

        tokens_per_element, consume_all = token_count(iparam)

        if consume_all:
            # Prepend the positional values to the keyword values.
            mapping.setdefault(iparam, [])
            pos_tokens = []

            for token in tokens:
                if not cparam.allow_leading_hyphen:
                    _validate_is_not_option_like(token)
                pos_tokens.append(token)
            mapping[iparam] = pos_tokens + mapping[iparam]
            tokens = []
            break

        tokens_per_element = max(1, tokens_per_element)

        if len(tokens) < tokens_per_element:
            raise MissingArgumentError(parameter=iparam, tokens_so_far=tokens)

        mapping.setdefault(iparam, [])
        for token in tokens[:tokens_per_element]:
            if not cparam.allow_leading_hyphen:
                _validate_is_not_option_like(token)

            mapping[iparam].append(token)

        tokens = tokens[tokens_per_element:]

    return tokens


def _parse_env(command: ResolvedCommand, mapping: ParameterDict):
    """Populate argument defaults from environment variables.

    In cyclopts, arguments are parsed with the following priority:

    1. CLI-provided values
    2. Values parsed from ``Parameter.env_var``.
    3. Default values from the function signature.
    """
    for iparam, cparam in command.iparam_to_cparam.items():
        if iparam in mapping:
            # Don't check environment variables for already-parsed parameters.
            continue

        for env_var_name in cparam.env_var:
            try:
                env_var_value = os.environ[env_var_name]
            except KeyError:
                pass
            else:
                mapping[iparam] = cparam.env_var_split(iparam.annotation, env_var_value)
                break


def _is_required(iparam: inspect.Parameter) -> bool:
    """A token must be provided for the given :class:``inspect.Parameter``."""
    return iparam.default is iparam.empty and iparam.kind not in (
        iparam.VAR_KEYWORD,
        iparam.VAR_POSITIONAL,
    )


def _bind(
    command: ResolvedCommand,
    mapping: ParameterDict,
):
    """Bind the mapping to the function signature.

    Better than directly using ``signature.bind`` because this can handle
    intermingled positional and keyword arguments.
    """
    f_pos, f_kwargs = [], {}
    use_pos = True

    def f_pos_append(p):
        nonlocal use_pos
        assert use_pos
        try:
            f_pos.append(mapping[p])
        except KeyError:
            use_pos = False

    for iparam in command.iparams:
        if use_pos and iparam.kind in (iparam.POSITIONAL_ONLY, iparam.POSITIONAL_OR_KEYWORD):
            f_pos_append(iparam)
        elif use_pos and iparam.kind is iparam.VAR_POSITIONAL:  # ``*args``
            f_pos.extend(mapping.get(iparam, []))
            use_pos = False
        elif iparam.kind is iparam.VAR_KEYWORD:
            f_kwargs.update(mapping.get(iparam, {}))
        else:
            with suppress(KeyError):
                f_kwargs[iparam.name] = mapping[iparam]

    bound = command.bind(*f_pos, **f_kwargs)
    return bound


def _convert(command: ResolvedCommand, mapping: ParameterDict) -> ParameterDict:
    coerced = ParameterDict()
    for iparam, parameter_tokens in mapping.items():
        cparam = command.iparam_to_cparam[iparam]
        type_ = get_hint_parameter(iparam)[0]

        # Checking if parameter_token is a string is a little jank,
        # but works for all current use-cases.
        for parameter_token in parameter_tokens:
            if not isinstance(parameter_token, str):
                # A token would be non-string if it's the implied-value (from a flag).
                coerced[iparam] = parameter_token
                break
        else:
            try:
                if iparam.kind == iparam.VAR_KEYWORD:
                    coerced[iparam] = {}
                    for key, values in parameter_tokens.items():
                        val = cparam.converter(type_, *values)
                        for validator in cparam.validator:
                            validator(type_, val)
                        coerced[iparam][key] = val
                elif iparam.kind == iparam.VAR_POSITIONAL:
                    val = cparam.converter(List[type_], *parameter_tokens)
                    for validator in cparam.validator:
                        for v in val:
                            validator(type_, v)
                    coerced[iparam] = val
                else:
                    val = cparam.converter(type_, *parameter_tokens)
                    for validator in cparam.validator:
                        validator(type_, val)
                    coerced[iparam] = val
            except CoercionError as e:
                e.parameter = iparam
                raise
            except (AssertionError, ValueError, TypeError) as e:
                raise ValidationError(value=e.args[0] if e.args else "", parameter=iparam) from e
    return coerced


def _walk_name_iparam_implicit_value(command: ResolvedCommand):
    for name, (iparam, implicit_value) in command.cli2parameter.items():
        if not name.startswith("--"):
            continue
        name = name[2:]  # Strip off the leading "--"
        yield name, iparam, implicit_value


def _parse_configs(command: ResolvedCommand, mapping: ParameterDict, configs):
    """Iteratively apply each ``config`` callable to the token mapping."""
    # Remap `mapping` back to CLI values for config parsing.
    cli_kwargs: Dict[str, Union[Unset, list]] = {}
    for cli_name, iparam, implicit_value in _walk_name_iparam_implicit_value(command):
        # Assign existing tokens to the "positive" flag/keyword if it exists.
        # Otherwise, assign it to the "negative" flag/keyword.
        cparam = command.iparam_to_cparam[iparam]
        if implicit_value is None or cparam.name == ("",) or implicit_value:
            with suppress(KeyError):
                cli_kwargs[cli_name] = mapping[iparam]

    def repopulate_unset():
        # Repopulate deleted keys with ``Unset``
        for name, iparam, _ in _walk_name_iparam_implicit_value(command):
            if name not in cli_kwargs or not cli_kwargs[name]:
                cli_kwargs[name] = Unset(iparam, {x[2:] for x in command.parameter2cli[iparam] if x.startswith("--")})

    repopulate_unset()

    for config in configs:
        config(cli_kwargs)
        repopulate_unset()

        # Validate that ``config`` produced reasonable modifications.
        # If there is an error at this stage, it is a developer-error of the config object
        for cli_name, values in cli_kwargs.items():
            if not isinstance(cli_name, str):
                raise TypeError(f"{config.func!r} produced non-str key {cli_name!r}.")
            if isinstance(values, Unset):
                continue
            if not isinstance(values, list):
                raise TypeError(f"{config.func!r} produced non-list value for key {cli_name!r}.")
            if "--" + cli_name not in command.cli2parameter:
                raise ValueError(f"{config.func!r} produced unknown key {cli_name!r}.")
            if len(values) > 1:
                # They all must be strings, or reinterpret as a list-of-list
                for value in values:
                    if not isinstance(value, str):
                        raise TypeError(f"{config.func!r} produced non-str element value for key {cli_name!r}.")

    # Rebind updated values to ``mapping``
    set_iparams = set()
    for cli_name, value in cli_kwargs.items():
        iparam, _ = command.cli2parameter["--" + cli_name]
        if isinstance(value, Unset):
            if not value.related_set(cli_kwargs):
                # No other "aliases" have provided values, safe to delete.
                # This can occur if a config Unsets/deletes a value.
                mapping.pop(iparam, None)
        elif id(iparam) in set_iparams and value != mapping[iparam]:
            # Intended to detect if a config sets different
            # values to different aliases of the same parameter.
            raise RepeatArgumentError(parameter=iparam)
        else:
            mapping[iparam] = value
            set_iparams.add(id(iparam))


def create_bound_arguments(
    command: ResolvedCommand,
    tokens: List[str],
    configs: Iterable[Callable],
) -> Tuple[inspect.BoundArguments, List[str]]:
    """Parse and coerce CLI tokens to match a function's signature.

    Parameters
    ----------
    command: ResolvedCommand
    tokens: List[str]
        CLI tokens to parse and coerce to match ``f``'s signature.

    Returns
    -------
    bound: inspect.BoundArguments
        The converted and bound positional and keyword arguments for ``f``.

    unused_tokens: List[str]
        Remaining tokens that couldn't be matched to ``f``'s signature.
    """
    # ``mapping`` maps inspect.Parameter to list of tokens/values.
    #    * Each token is USUALLY a string and needs further casting/interpretation.
    #    * However, if it's NOT a string, the value should be used as-is.
    #        * This is used for implicit-value keyword tokens like "--flag" and "--empty-iterable"
    # ``mapping`` is updated inplace throughout this function.
    mapping = ParameterDict()
    unused_tokens = []

    validate_command(command.command)

    try:
        # Build up a mapping of inspect.Parameter->List[str]
        unused_tokens = _parse_kw_and_flags(command, tokens, mapping)
        unused_tokens = _parse_pos(command, unused_tokens, mapping)
        _parse_env(command, mapping)
        _parse_configs(command, mapping, configs)

        # For each parameter, convert the list of string tokens.
        coerced = _convert(command, mapping)
        bound = _bind(command, coerced)

        # Apply group converters
        for group, iparams in command.groups_iparams:
            if not group.converter:
                continue
            names = tuple(x.name for x in iparams)
            converted = group.converter(**{k: bound.arguments[k] for k in names if k in bound.arguments})
            for name in names:  # Merge back in the result
                try:
                    bound.arguments[name] = converted[name]
                except KeyError:
                    bound.arguments.pop(name, None)

        # Apply group validators
        try:
            for group, iparams in command.groups_iparams:
                names = tuple(x.name for x in iparams)
                for validator in group.validator:  # pyright: ignore
                    validator(**{k: bound.arguments[k] for k in names if k in bound.arguments})
        except (AssertionError, ValueError, TypeError) as e:
            # group will always be set from the above for loop if an exception occurs.
            raise ValidationError(
                value=e.args[0] if e.args else "",
                group=group,  # pyright: ignore[reportPossiblyUnboundVariable]
            ) from e

        for iparam in command.iparams:
            if _is_required(iparam) and iparam.name not in bound.arguments:
                raise MissingArgumentError(parameter=iparam)

    except CycloptsError as e:
        e.target = command.command
        e.root_input_tokens = tokens
        e.cli2parameter = command.cli2parameter
        e.parameter2cli = command.parameter2cli
        e.unused_tokens = unused_tokens
        raise

    return bound, unused_tokens



import inspect
import os
import sys
import traceback
from contextlib import suppress
from copy import copy
from functools import partial
from itertools import chain
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Literal,
    Mapping,
    Optional,
    Tuple,
    TypeVar,
    Union,
    overload,
)

from attrs import define, field

import cyclopts.utils
from cyclopts.bind import create_bound_arguments, normalize_tokens
from cyclopts.exceptions import (
    CommandCollisionError,
    CycloptsError,
    InvalidCommandError,
    UnusedCliTokensError,
    ValidationError,
    format_cyclopts_error,
)
from cyclopts.group import Group, GroupConverter, sort_groups
from cyclopts.group_extractors import groups_from_app, inverse_groups_from_app
from cyclopts.help import (
    HelpPanel,
    create_parameter_help_panel,
    format_command_entries,
    format_doc,
    format_str,
    format_usage,
    resolve_help_format,
    resolve_version_format,
)
from cyclopts.parameter import Parameter, validate_command
from cyclopts.protocols import Dispatcher
from cyclopts.resolve import ResolvedCommand
from cyclopts.utils import (
    default_name_transform,
    optional_to_tuple_converter,
    to_list_converter,
    to_tuple_converter,
)

T = TypeVar("T", bound=Callable)


if sys.version_info < (3, 9):
    from typing_extensions import Annotated
else:
    from typing import Annotated


with suppress(ImportError):
    # By importing, makes things like the arrow-keys work.
    import readline  # Not available on windows

if TYPE_CHECKING:
    from rich.console import Console


class _CannotDeriveCallingModuleNameError(Exception):
    pass


def _get_root_module_name():
    """Get the calling package name from the call-stack."""
    for elem in inspect.stack():
        module = inspect.getmodule(elem.frame)
        if module is None:
            continue
        root_module_name = module.__name__.split(".")[0]
        if root_module_name == "cyclopts":
            continue
        return root_module_name

    raise _CannotDeriveCallingModuleNameError  # pragma: no cover


def _default_version(default="0.0.0") -> str:
    """Attempts to get the calling code's version.

    Returns
    -------
    version: str
        ``default`` if it cannot determine version.
    """
    import importlib

    if sys.version_info < (3, 10):  # pragma: no cover
        from importlib_metadata import PackageNotFoundError
        from importlib_metadata import version as importlib_metadata_version
    else:  # pragma: no cover
        from importlib.metadata import PackageNotFoundError
        from importlib.metadata import version as importlib_metadata_version

    try:
        root_module_name = _get_root_module_name()
    except _CannotDeriveCallingModuleNameError:  # pragma: no cover
        return default

    # Attempt to get the Distribution Package’s version number.
    try:
        return importlib_metadata_version(root_module_name)
    except PackageNotFoundError:
        pass

    # Attempt packagename.__version__
    # Not sure if this is redundant with ``importlib.metadata``,
    # but there's no real harm in checking.
    try:
        module = importlib.import_module(root_module_name)
        return module.__version__
    except (ImportError, AttributeError):
        pass

    # Final fallback
    return default


def _validate_default_command(x):
    if isinstance(x, App):
        raise TypeError("Cannot register a sub-App to default.")
    return x


def _combined_meta_command_mapping(app):
    """Return a copied and combined mapping containing app and meta-app commands."""
    command_mapping = copy(app._commands)
    while (app := app._meta) and app._commands:
        command_mapping.update(app._commands)
    return command_mapping


def _get_command_groups(parent_app, child_app):
    """Extract out the command groups from the ``parent_app`` for a given ``child_app``."""
    return next(x for x in inverse_groups_from_app(parent_app) if x[0] is child_app)[1]


def resolve_default_parameter_from_apps(apps) -> Parameter:
    """The default_parameter resolution depends on the parent-child path traversed."""
    cparams = []
    for parent_app, child_app in zip(apps[:-1], apps[1:]):
        # child_app could be a command of parent_app.meta
        if parent_app._meta and child_app in parent_app._meta._commands.values():
            cparams = []  # meta-apps do NOT inherit from their parenting app.
            parent_app = parent_app._meta

        groups = _get_command_groups(parent_app, child_app)
        cparams.extend([group.default_parameter for group in groups])
        cparams.append(parent_app.default_parameter)

    cparams.append(apps[-1].default_parameter)

    return Parameter.combine(*cparams)


def walk_metas(app):
    # Iterates from deepest to shallowest meta-apps
    meta_list = [app]  # shallowest to deepest
    meta = app
    while (meta := meta._meta) and meta.default_command:
        meta_list.append(meta)
    yield from reversed(meta_list)


@define
class App:
    # This can ONLY ever be Tuple[str, ...] due to converter.
    # The other types is to make mypy happy for Cyclopts users.
    _name: Union[None, str, Tuple[str, ...]] = field(default=None, alias="name", converter=optional_to_tuple_converter)

    _help: Optional[str] = field(default=None, alias="help")

    usage: Optional[str] = field(default=None)

    # Everything below must be kw_only

    default_command: Optional[Callable] = field(default=None, converter=_validate_default_command, kw_only=True)
    default_parameter: Optional[Parameter] = field(default=None, kw_only=True)

    # This can ONLY ever be None or Tuple[Callable, ...]
    _config: Union[None, Callable, Iterable[Callable]] = field(
        default=None,
        alias="config",
        converter=optional_to_tuple_converter,
        kw_only=True,
    )

    version: Union[None, str, Callable[..., str]] = field(default=_default_version, kw_only=True)
    # This can ONLY ever be a Tuple[str, ...]
    _version_flags: Union[str, Iterable[str]] = field(
        default=["--version"],
        converter=to_tuple_converter,
        alias="version_flags",
        kw_only=True,
    )

    show: bool = field(default=True, kw_only=True)

    console: Optional["Console"] = field(default=None, kw_only=True)

    # This can ONLY ever be a Tuple[str, ...]
    _help_flags: Union[str, Iterable[str]] = field(
        default=["--help", "-h"],
        converter=to_tuple_converter,
        alias="help_flags",
        kw_only=True,
    )
    help_format: Optional[
        Literal[
            "markdown",
            "md",
            "plaintext",
            "restructuredtext",
            "rst",
            "rich",
        ]
    ] = field(default=None, kw_only=True)

    version_format: Optional[
        Literal[
            "markdown",
            "md",
            "plaintext",
            "restructuredtext",
            "rst",
            "rich",
        ]
    ] = field(default=None, kw_only=True)

    # This can ONLY ever be Tuple[Union[Group, str], ...] due to converter.
    # The other types is to make mypy happy for Cyclopts users.
    group: Union[Group, str, Tuple[Union[Group, str], ...]] = field(
        default=None, converter=to_tuple_converter, kw_only=True
    )

    # This can ONLY ever be a Group
    group_arguments: Union[Group, str, None] = field(
        default=None,
        converter=GroupConverter(Group.create_default_arguments()),
        kw_only=True,
    )
    # This can ONLY ever be a Group
    group_parameters: Union[Group, str, None] = field(
        default=None,
        converter=GroupConverter(Group.create_default_parameters()),
        kw_only=True,
    )
    # This can ONLY ever be a Group
    group_commands: Union[Group, str, None] = field(
        default=None,
        converter=GroupConverter(Group.create_default_commands()),
        kw_only=True,
    )

    converter: Optional[Callable[..., Mapping[str, Any]]] = field(default=None, kw_only=True)
    validator: List[Callable[..., Any]] = field(default=None, converter=to_list_converter, kw_only=True)

    _name_transform: Optional[Callable[[str], str]] = field(
        default=None,
        alias="name_transform",
        kw_only=True,
    )

    ######################
    # Private Attributes #
    ######################
    # Maps CLI-name of a command to a function handle.
    _commands: Dict[str, "App"] = field(init=False, factory=dict)

    _parents: List["App"] = field(init=False, factory=list)

    _meta: "App" = field(init=False, default=None)
    _meta_parent: "App" = field(init=False, default=None)

    def __attrs_post_init__(self):
        # Trigger the setters
        self.help_flags = self._help_flags
        self.version_flags = self._version_flags

    ###########
    # Methods #
    ###########
    def _delete_commands(self, commands: Iterable[str], default=None):
        """Safely delete commands.

        Will **not** raise an exception if command(s) do not exist.

        Parameters
        ----------
        commands: Iterable[str, ...]
            Strings of commands to delete.
        """
        # Remove all the old version-flag commands.
        for command in commands:
            with suppress(KeyError):
                if default:
                    if self[command].default == default:
                        del self[command]
                else:
                    del self[command]

    @property
    def version_flags(self):
        return self._version_flags

    @version_flags.setter
    def version_flags(self, value):
        self._version_flags = value
        self._delete_commands(self._version_flags, default=self.version_print)
        if self._version_flags:
            self.command(
                self.version_print,
                name=self._version_flags,
                help_flags=[],
                version_flags=[],
                version=self.version,
                help="Display application version.",
            )

    @property
    def help_flags(self):
        return self._help_flags

    @help_flags.setter
    def help_flags(self, value):
        self._help_flags = value
        self._delete_commands(self._help_flags, default=self.help_print)
        if self._help_flags:
            self.command(
                self.help_print,
                name=self._help_flags,
                help_flags=[],
                version_flags=[],
                version=self.version,
                help="Display this message and exit.",
            )

    @property
    def name(self) -> Tuple[str, ...]:
        """Application name(s). Dynamically derived if not previously set."""
        if self._name:
            return self._name  # pyright: ignore[reportReturnType]
        elif self.default_command is None:
            name = Path(sys.argv[0]).name
            if name == "__main__.py":
                name = _get_root_module_name()
            return (name,)
        else:
            return (self.name_transform(self.default_command.__name__),)

    @property
    def config(self) -> Tuple[str, ...]:
        if self._config is None and self._meta_parent is not None:
            return self._meta_parent.config
        else:
            return self._config or ()  # pyright: ignore[reportReturnType]

    @config.setter
    def config(self, value):
        self._config = value

    @property
    def help(self) -> str:
        if self._help is not None:
            return self._help
        elif self.default_command is None:
            # Try and fallback to a meta-app docstring.
            if self._meta is None:
                return ""
            else:
                return self.meta.help
        elif self.default_command.__doc__ is None:
            return ""
        else:
            return self.default_command.__doc__

    @help.setter
    def help(self, value):
        self._help = value

    @property
    def name_transform(self):
        return self._name_transform if self._name_transform else default_name_transform

    @name_transform.setter
    def name_transform(self, value):
        self._name_transform = value

    def version_print(
        self,
        console: Optional["Console"] = None,
    ) -> None:
        """Print the application version.

        Parameters
        ----------
        console: rich.console.Console
            Console to print version string to.
            If not provided, follows the resolution order defined in :attr:`App.console`.

        """
        console = self._resolve_console(None, console)
        version_format = resolve_version_format([self])

        version_raw = self.version() if callable(self.version) else self.version

        if version_raw is None:
            version_raw = "0.0.0"

        version_formatted = format_str(version_raw, format=version_format)
        console.print(version_formatted)

    def __getitem__(self, key: str) -> "App":
        """Get the subapp from a command string.

        All commands get registered to Cyclopts as subapps.
        The actual function handler is at ``app[key].default_command``.
        """
        if self._meta:
            with suppress(KeyError):
                return self.meta[key]
        return self._commands[key]

    def __delitem__(self, key: str):
        del self._commands[key]

    def __contains__(self, k: str) -> bool:
        if k in self._commands:
            return True
        if self._meta_parent:
            return k in self._meta_parent
        return False

    def __iter__(self) -> Iterator[str]:
        """Iterate over command & meta command names."""
        for k in self._commands:
            yield k
        if self._meta_parent:
            for k in self._meta_parent:
                yield k

    @property
    def meta(self) -> "App":
        if self._meta is None:
            self._meta = type(self)(
                help_flags=self.help_flags,
                version_flags=self.version_flags,
                group_commands=copy(self.group_commands),
                group_arguments=copy(self.group_arguments),
                group_parameters=copy(self.group_parameters),
            )
            self._meta._meta_parent = self
        return self._meta

    def parse_commands(
        self,
        tokens: Union[None, str, Iterable[str]] = None,
    ) -> Tuple[Tuple[str, ...], Tuple["App", ...], List[str]]:
        """Extract out the command tokens from a command.

        Parameters
        ----------
        tokens: Union[None, str, Iterable[str]]
            Either a string, or a list of strings to launch a command.
            Defaults to ``sys.argv[1:]``

        Returns
        -------
        List[str]
            Tokens that are interpreted as a valid command chain.
        List[App]
            The associated :class:`App` object with each of those tokens.
        List[str]
            The remaining non-command tokens.
        """
        tokens = normalize_tokens(tokens)

        command_chain = []
        app = self
        apps = [app]
        unused_tokens = tokens

        command_mapping = _combined_meta_command_mapping(app)

        for i, token in enumerate(tokens):
            try:
                app = command_mapping[token]
                apps.append(app)
                unused_tokens = tokens[i + 1 :]
            except KeyError:
                break
            command_chain.append(token)
            command_mapping = _combined_meta_command_mapping(app)

        return tuple(command_chain), tuple(apps), unused_tokens

    # This overload is used in code like:
    #
    # @app.command
    # def my_command(foo: str):
    #   ...
    @overload
    def command(
        self,
        obj: T,
        name: Union[None, str, Iterable[str]] = None,
        **kwargs: object,
    ) -> T: ...

    # This overload is used in code like:
    #
    # @app.command(name="bar")
    # def my_command(foo: str):
    #   ...
    @overload
    def command(
        self,
        obj: None = None,
        name: Union[None, str, Iterable[str]] = None,
        **kwargs: object,
    ) -> Callable[[T], T]: ...

    def command(
        self,
        obj: Optional[T] = None,
        name: Union[None, str, Iterable[str]] = None,
        **kwargs: object,
    ) -> Union[T, Callable[[T], T]]:
        """Decorator to register a function as a CLI command.

        Parameters
        ----------
        obj: Optional[Callable]
            Function or :class:`App` to be registered as a command.
        name: Union[None, str, Iterable[str]]
            Name(s) to register the ``obj`` to.
            If not provided, defaults to:

            * If registering an :class:`App`, then the app's name.
            * If registering a function, then the function's name.
        `**kwargs`
            Any argument that :class:`App` can take.
        """
        if obj is None:  # Called ``@app.command(...)``
            return partial(self.command, name=name, **kwargs)  # pyright: ignore[reportReturnType]

        if isinstance(obj, App):
            app = obj

            if app._name is None and name is None:
                raise ValueError("Sub-app MUST have a name specified.")

            if kwargs:
                raise ValueError("Cannot supplied additional configuration when registering a sub-App.")
        else:
            validate_command(obj)

            kwargs.setdefault("help_flags", self.help_flags)
            kwargs.setdefault("version_flags", self.version_flags)

            if "group_commands" not in kwargs:
                kwargs["group_commands"] = copy(self.group_commands)
            if "group_parameters" not in kwargs:
                kwargs["group_parameters"] = copy(self.group_parameters)
            if "group_arguments" not in kwargs:
                kwargs["group_arguments"] = copy(self.group_arguments)
            app = App(default_command=obj, **kwargs)  # pyright: ignore

            for flag in chain(kwargs["help_flags"], kwargs["version_flags"]):  # pyright: ignore
                app[flag].show = False

        if app._name_transform is None:
            app.name_transform = self.name_transform

        if name is None:
            name = app.name
        else:
            app._name = name  # pyright: ignore[reportAttributeAccessIssue]

        for n in to_tuple_converter(name):
            if n in self:
                raise CommandCollisionError(f'Command "{n}" already registered.')

            # Warning: app._name may not align with command name
            self._commands[n] = app

        app._parents.append(self)

        return obj  # pyright: ignore[reportReturnType]

    # This overload is used in code like:
    #
    # @app.default
    # def my_command(foo: str):
    #   ...
    @overload
    def default(
        self,
        obj: T,
        *,
        converter: Optional[Callable[..., Mapping[str, Any]]] = None,
        validator: Optional[Callable[..., Any]] = None,
    ) -> T: ...

    # This overload is used in code like:
    #
    # @app.default()
    # def my_command(foo: str):
    #   ...
    @overload
    def default(
        self,
        obj: None = None,
        *,
        converter: Optional[Callable[..., Mapping[str, Any]]] = None,
        validator: Optional[Callable[..., Any]] = None,
    ) -> Callable[[T], T]: ...

    def default(
        self,
        obj: Optional[T] = None,
        *,
        converter: Optional[Callable[..., Mapping[str, Any]]] = None,
        validator: Optional[Callable[..., Any]] = None,
    ) -> Union[T, Callable[[T], T]]:
        """Decorator to register a function as the default action handler."""
        if obj is None:  # Called ``@app.default_command(...)``
            return partial(self.default, converter=converter, validator=validator)  # pyright: ignore[reportReturnType]

        if isinstance(obj, App):  # Registering a sub-App
            raise TypeError("Cannot register a sub-App to default.")

        if self.default_command is not None:
            raise CommandCollisionError(f"Default command previously set to {self.default_command}.")

        validate_command(obj)
        self.default_command = obj
        if converter:
            self.converter = converter
        if validator:
            self.validator = validator  # pyright: ignore[reportAttributeAccessIssue]
        return obj

    def parse_known_args(
        self,
        tokens: Union[None, str, Iterable[str]] = None,
        *,
        console: Optional["Console"] = None,
    ) -> Tuple[Callable, inspect.BoundArguments, List[str]]:
        """Interpret arguments into a function, :class:`~inspect.BoundArguments`, and any remaining unknown tokens.

        Parameters
        ----------
        tokens: Union[None, str, Iterable[str]]
            Either a string, or a list of strings to launch a command.
            Defaults to ``sys.argv[1:]``
        console: rich.console.Console
            Console to print help and runtime Cyclopts errors.
            If not provided, follows the resolution order defined in :attr:`App.console`.

        Returns
        -------
        command: Callable
            Bare function to execute.

        bound: inspect.BoundArguments
            Bound arguments for ``command``.

        unused_tokens: List[str]
            Any remaining CLI tokens that didn't get parsed for ``command``.
        """
        tokens = normalize_tokens(tokens)

        meta_parent = self

        command_chain, apps, unused_tokens = self.parse_commands(tokens)
        command_app = apps[-1]

        # We don't want the command_app to be the version/help handler.
        with suppress(IndexError):
            if set(command_app.name) & set(apps[-2].help_flags + apps[-2].version_flags):  # pyright: ignore
                apps = apps[:-1]
                command_app = apps[-1]

        try:
            parent_app = apps[-2]
        except IndexError:
            parent_app = None

        config: Tuple[Callable, ...] = ()
        for app in reversed(apps):
            if app.config:
                config = app.config  # pyright: ignore[reportAssignmentType]
                break
        config = tuple(partial(x, apps, command_chain) for x in config)

        # Special flags (help/version) get intercepted by the root app.
        # Special flags are allows to be **anywhere** in the token stream.

        for help_flag in command_app.help_flags:
            try:
                help_flag_index = tokens.index(help_flag)
                break
            except ValueError:
                pass
        else:
            help_flag_index = None

        if help_flag_index is not None:
            tokens.pop(help_flag_index)
            command = self.help_print
            while meta_parent := meta_parent._meta_parent:
                command = meta_parent.help_print
            bound = cyclopts.utils.signature(command).bind(tokens, console=console)
            unused_tokens = []
        elif any(flag in tokens for flag in command_app.version_flags):
            # Version
            command = self.version_print
            while meta_parent := meta_parent._meta_parent:
                command = meta_parent.version_print
            bound = cyclopts.utils.signature(command).bind()
            unused_tokens = []
        else:
            try:
                if command_app.default_command:
                    command = command_app.default_command
                    resolved_command = self._resolve_command(tokens, parse_docstring=False)
                    # We want the resolved group that ``app`` belongs to.
                    command_groups = [] if parent_app is None else _get_command_groups(parent_app, command_app)

                    bound, unused_tokens = create_bound_arguments(resolved_command, unused_tokens, config)
                    try:
                        if command_app.converter:
                            bound.arguments = command_app.converter(**bound.arguments)  # pyright: ignore[reportAttributeAccessIssue]
                        for command_group in command_groups:
                            if command_group.converter:
                                bound.arguments = command_group.converter(**bound.arguments)
                        for validator in command_app.validator:
                            validator(**bound.arguments)
                        for command_group in command_groups:
                            for validator in command_group.validator:  # pyright: ignore
                                validator(**bound.arguments)
                    except (AssertionError, ValueError, TypeError) as e:
                        raise ValidationError(
                            value=e.args[0] if e.args else "",
                            group=command_group,  # pyright: ignore[reportPossiblyUnboundVariable]
                        ) from e

                else:
                    if unused_tokens:
                        raise InvalidCommandError(unused_tokens=unused_tokens)
                    else:
                        # Running the application with no arguments and no registered
                        # ``default_command`` will default to ``help_print``.
                        command = self.help_print
                        bound = cyclopts.utils.signature(command).bind(tokens=tokens, console=console)
                        unused_tokens = []
            except CycloptsError as e:
                e.app = command_app
                if command_chain:
                    e.command_chain = command_chain
                if e.console is None:
                    e.console = self._resolve_console(tokens, console)
                raise

        return command, bound, unused_tokens

    def parse_args(
        self,
        tokens: Union[None, str, Iterable[str]] = None,
        *,
        console: Optional["Console"] = None,
        print_error: bool = True,
        exit_on_error: bool = True,
        verbose: bool = False,
    ) -> Tuple[Callable, inspect.BoundArguments]:
        """Interpret arguments into a function and :class:`~inspect.BoundArguments`.

        Raises
        ------
        UnusedCliTokensError
            If any tokens remain after parsing.

        Parameters
        ----------
        tokens: Union[None, str, Iterable[str]]
            Either a string, or a list of strings to launch a command.
            Defaults to ``sys.argv[1:]``.
        console: rich.console.Console
            Console to print help and runtime Cyclopts errors.
            If not provided, follows the resolution order defined in :attr:`App.console`.
        print_error: bool
            Print a rich-formatted error on error.
            Defaults to ``True``.
        exit_on_error: bool
            If there is an error parsing the CLI tokens invoke ``sys.exit(1)``.
            Otherwise, continue to raise the exception.
            Defaults to ``True``.
        verbose: bool
            Populate exception strings with more information intended for developers.
            Defaults to ``False``.

        Returns
        -------
        command: Callable
            Function associated with command action.

        bound: inspect.BoundArguments
            Parsed and converted ``args`` and ``kwargs`` to be used when calling ``command``.
        """
        tokens = normalize_tokens(tokens)

        # Normal parsing
        try:
            command, bound, unused_tokens = self.parse_known_args(tokens, console=console)
            if unused_tokens:
                raise UnusedCliTokensError(
                    target=command,
                    unused_tokens=unused_tokens,
                )
        except CycloptsError as e:
            e.verbose = verbose
            e.root_input_tokens = tokens

            if e.console is None:
                e.console = self._resolve_console(tokens, console)
            if print_error:
                assert e.console
                e.console.print(format_cyclopts_error(e))
            if exit_on_error:
                sys.exit(1)
            raise

        return command, bound

    def __call__(
        self,
        tokens: Union[None, str, Iterable[str]] = None,
        *,
        console: Optional["Console"] = None,
        print_error: bool = True,
        exit_on_error: bool = True,
        verbose: bool = False,
    ):
        """Interprets and executes a command.

        Parameters
        ----------
        tokens : Union[None, str, Iterable[str]]
            Either a string, or a list of strings to launch a command.
            Defaults to ``sys.argv[1:]``.
        console: rich.console.Console
            Console to print help and runtime Cyclopts errors.
            If not provided, follows the resolution order defined in :attr:`App.console`.
        print_error: bool
            Print a rich-formatted error on error.
            Defaults to ``True``.
        exit_on_error: bool
            If there is an error parsing the CLI tokens invoke ``sys.exit(1)``.
            Otherwise, continue to raise the exception.
            Defaults to ``True``.
        verbose: bool
            Populate exception strings with more information intended for developers.
            Defaults to ``False``.

        Returns
        -------
        return_value: Any
            The value the parsed command handler returns.
        """
        tokens = normalize_tokens(tokens)
        command, bound = self.parse_args(
            tokens,
            console=console,
            print_error=print_error,
            exit_on_error=exit_on_error,
            verbose=verbose,
        )
        try:
            if inspect.iscoroutinefunction(command):
                import asyncio

                return asyncio.run(command(*bound.args, **bound.kwargs))
            else:
                return command(*bound.args, **bound.kwargs)
        except Exception as e:
            try:
                from pydantic import ValidationError as PydanticValidationError
            except ImportError:
                PydanticValidationError = None  # noqa: N806

            if PydanticValidationError is not None and isinstance(e, PydanticValidationError):
                if print_error:
                    console = self._resolve_console(tokens, console)
                    console.print(format_cyclopts_error(e))

                if exit_on_error:
                    sys.exit(1)
            raise

    def _resolve_console(
        self, tokens: Union[None, str, Iterable[str]], console: Optional["Console"] = None
    ) -> "Console":
        if console is not None:
            return console
        _, apps, _ = self.parse_commands(tokens)
        for app in reversed(apps):
            if app.console:
                return app.console
        from rich.console import Console

        return Console()

    def help_print(
        self,
        tokens: Annotated[Union[None, str, Iterable[str]], Parameter(show=False)] = None,
        *,
        console: Annotated[Optional["Console"], Parameter(parse=False)] = None,
    ) -> None:
        """Print the help page.

        Parameters
        ----------
        tokens: Union[None, str, Iterable[str]]
            Tokens to interpret for traversing the application command structure.
            If not provided, defaults to ``sys.argv``.
        console: rich.console.Console
            Console to print help and runtime Cyclopts errors.
            If not provided, follows the resolution order defined in :attr:`App.console`.
        """
        tokens = normalize_tokens(tokens)

        command_chain, apps, _ = self.parse_commands(tokens)
        executing_app = apps[-1]

        console = self._resolve_console(tokens, console)

        # Print the:
        #    my-app command COMMAND [ARGS] [OPTIONS]
        if executing_app.usage is None:
            console.print(format_usage(self, command_chain))
        elif executing_app.usage:  # i.e. skip empty-string.
            console.print(executing_app.usage + "\n")

        # Print the App/Command's Doc String.
        help_format = resolve_help_format(apps)
        console.print(format_doc(self, executing_app, help_format))

        for help_panel in self._assemble_help_panels(tokens, help_format):
            console.print(help_panel)

    def _resolve_command(
        self,
        tokens: Union[None, str, Iterable[str]] = None,
        parse_docstring: bool = True,
    ) -> ResolvedCommand:
        _, apps, _ = self.parse_commands(tokens)

        if not apps[-1].default_command:
            raise InvalidCommandError

        assert isinstance(apps[-1].group_arguments, Group)
        assert isinstance(apps[-1].group_parameters, Group)

        resolved_command = ResolvedCommand(
            apps[-1].default_command,
            resolve_default_parameter_from_apps(apps),
            apps[-1].group_arguments,
            apps[-1].group_parameters,
            parse_docstring=parse_docstring,
        )
        return resolved_command

    def _assemble_help_panels(
        self,
        tokens: Union[None, str, Iterable[str]],
        help_format,
    ) -> List[HelpPanel]:
        from rich.console import Group as RichGroup
        from rich.console import NewLine

        _, apps, _ = self.parse_commands(tokens)

        help_format = resolve_help_format(apps)

        panels: Dict[str, Tuple[Group, HelpPanel]] = {}
        # Handle commands first; there's an off chance they may be "upgraded"
        # to an argument/parameter panel.
        for subapp in walk_metas(apps[-1]):
            # Handle Commands
            for group, elements in groups_from_app(subapp):
                if not group.show:
                    continue

                try:
                    _, command_panel = panels[group.name]
                except KeyError:
                    command_panel = HelpPanel(
                        format="command",
                        title=group.name,
                    )
                    panels[group.name] = (group, command_panel)

                if group.help:
                    group_help = format_str(group.help, format=help_format)

                    if command_panel.description:
                        command_panel.description = RichGroup(command_panel.description, NewLine(), group_help)
                    else:
                        command_panel.description = group_help

                command_panel.entries.extend(format_command_entries(elements, format=help_format))

        # Handle Arguments/Parameters
        for subapp in walk_metas(apps[-1]):
            if not subapp.default_command:
                continue
            command = ResolvedCommand(
                subapp.default_command,
                resolve_default_parameter_from_apps(apps),
                subapp.group_arguments,
                subapp.group_parameters,
            )
            for group, iparams in command.groups_iparams:
                if not group.show:
                    continue
                cparams = [command.iparam_to_cparam[x] for x in iparams]
                try:
                    _, existing_panel = panels[group.name]
                except KeyError:
                    existing_panel = None
                new_panel = create_parameter_help_panel(group, iparams, cparams, help_format)

                if existing_panel:
                    # An imperfect merging process
                    existing_panel.format = "parameter"
                    existing_panel.entries = new_panel.entries + existing_panel.entries  # Commands go last
                    if new_panel.description:
                        if existing_panel.description:
                            existing_panel.description = RichGroup(
                                existing_panel.description, NewLine(), new_panel.description
                            )
                        else:
                            existing_panel.description = new_panel.description
                else:
                    panels[group.name] = (group, new_panel)

        groups = [x[0] for x in panels.values()]
        help_panels = [x[1] for x in panels.values()]

        out = []
        for help_panel in sort_groups(groups, help_panels)[1]:
            help_panel.remove_duplicates()
            if help_panel.format == "command":
                # don't sort format == "parameter" because order may matter there!
                help_panel.sort()
            out.append(help_panel)
        return out

    def interactive_shell(
        self,
        prompt: str = "$ ",
        quit: Union[None, str, Iterable[str]] = None,
        dispatcher: Optional[Dispatcher] = None,
        **kwargs,
    ) -> None:
        """Create a blocking, interactive shell.

        All registered commands can be executed in the shell.

        Parameters
        ----------
        prompt: str
            Shell prompt. Defaults to ``"$ "``.
        quit: Union[str, Iterable[str]]
            String or list of strings that will cause the shell to exit and this method to return.
            Defaults to ``["q", "quit"]``.
        dispatcher: Optional[Dispatcher]
            Optional function that subsequently invokes the command.
            The ``dispatcher`` function must have signature:

            .. code-block:: python

                def dispatcher(command: Callable, bound: inspect.BoundArguments) -> Any:
                    return command(*bound.args, **bound.kwargs)

            The above is the default dispatcher implementation.
        `**kwargs`
            Get passed along to :meth:`parse_args`.
        """
        if os.name == "posix":
            print("Interactive shell. Press Ctrl-D to exit.")
        else:  # Windows
            print("Interactive shell. Press Ctrl-Z followed by Enter to exit.")

        if quit is None:
            quit = ["q", "quit"]
        if isinstance(quit, str):
            quit = [quit]

        def default_dispatcher(command, bound):
            return command(*bound.args, **bound.kwargs)

        if dispatcher is None:
            dispatcher = default_dispatcher

        kwargs.setdefault("exit_on_error", False)

        while True:
            try:
                user_input = input(prompt)
            except EOFError:
                break

            tokens = normalize_tokens(user_input)
            if not tokens:
                continue
            if tokens[0] in quit:
                break

            try:
                command, bound = self.parse_args(tokens, **kwargs)
                dispatcher(command, bound)
            except CycloptsError:
                # Upstream ``parse_args`` already printed the error
                pass
            except Exception:
                print(traceback.format_exc())

    def __repr__(self):
        """Only shows non-default values."""
        non_defaults = {}
        for a in self.__attrs_attrs__:  # pyright: ignore[reportAttributeAccessIssue]
            if not a.init:
                continue
            v = getattr(self, a.name)
            # Compare types first because of some weird attribute issues.
            if type(v) != type(a.default) or v != a.default:  # noqa: E721
                non_defaults[a.alias] = v

        signature = ", ".join(f"{k}={v!r}" for k, v in non_defaults.items())
        return f"{type(self).__name__}({signature})"
